#!/usr/bin/perl
#
# ipacct v. $vers
#
# Written by Tim Yardley [yardley@uiuc.edu] (1999) for Empire Net, Inc.
#
# 1.06, 1.08 Fixes from Bob Zoller [bob@empnet.com] 
#
# You are encouraged to send fixes/enhancements to ipacct@empnet.com to be
# included in a later version of this program.
#
# http://www2.empnet.com/ipacct/
#
# Copyright (C) 1998, 1999 Empire Net, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

use Time::localtime;

$vers='1.08';

###################################
#
# start modifications
#  - change these globals for your company/setup
#

$COMP_NAME = 'Empire Net';			# Company Name
$INTERFACE = 'fxp';				# Interface starting string
$IPF_LOC = '/etc/ipf.rules';			# Full path/filename of IPF rules

#
# end modifications
#
###################################

#
# defines - global
#
$LOCK_EX = 2;			# Exclusive
$LOCK_UN = 8;			# Unlock

$html = 'ipacct.html';		# ip accounting monthly report
$names = 'ipnames.db';		# ip to company names

$temp = (localtime)->mday . '-tacct.db';		# temp filename
$day = (localtime)->mday . '-dacct.db';			# day filename
$dayin = (localtime)->mday . '-diacct.db';		# day input filename
$dayout = (localtime)->mday . '-doacct.db';		# day output filename
$month = ((localtime)->mon + 1) . '-' . ((localtime)->year+1900) . '-macct.db';		# month filename
$currentdate = (localtime)->hour . ':' . (localtime)->min . ' - ' . ((localtime)->mon + 1) . '/' . (localtime)->mday . '/' . ((localtime)->year+1900);	# current date

#
# prints usage information
#
sub usage {
	print(":-\n");
	print(": IP Accounting v$vers\n");
	print(": syntax: $0 [-crsw]\n");
	print(":   -c : compile stats (daily and monthly)\n");
	print(":   -r : reset ip counters (reload $IPF_LOC)\n");
	print(":   -w [output html filename] [input db filename] : web page creation\n");
	print(":   -s <iprange> [output html filename] [input db filename] : web page creation\n");
	print(":                                      only for IP range from $names\n");
	print(": iprange should be specified as the first IP listed in a range of ips\n");
	print(": specified in $names\n");
	print(":-\n");
	exit;
}

#
# this simply locks the file for exclusive use.
#
sub lock {
	local($file) = @_;
	flock($file, $LOCK_EX);
}

#
# unlocks the file after you're done with it.
#
sub unlock {
	local($file) = @_;
	flock($file, $LOCK_UN);
}

#
# searches for ip information from monthly stats
#
sub search {
	my($pattern) = @_;
	open(NAMES, "<$names") || die("Unable to open $names: $!\n");
	&lock(NAMES);
	open(MONTH, "<$month") || die("Unable to open $month: $!\n");
	&lock(MONTH);
	open(HTML, ">$html") || die("Unable to open $html: $!\n");
	&lock(HTML);

	#
	# print headers
	#
	print (HTML '<html>' . "\n");
	print (HTML '<head>' . "\n");
	print (HTML '  <title>IP Accounting Information</title>' . "\n");
	print (HTML '</head>' . "\n");
	print (HTML '<body bgcolor="#ffffff">' . "\n");

	#
	# handle date headers
	#
	$line = <MONTH>;
	print (HTML $line);
	$line = <MONTH>;
	print (HTML "\nReport Date: $currentdate\n");
	
	#
	# begin table
	#
	print (HTML '<table border="0" width="90%">' . "\n");
	print (HTML '  <tr>' . "\n");
	print (HTML '    <td align="center"><b><u> IP address </u></b></td>' . "\n");
	print (HTML '    <td align="center"><b><u> MBytes in | GBytes in </u></b></td>' . "\n");
	print (HTML '    <td align="center"><b><u> MBytes out | GBytes out </u></b></td>' . "\n");
	print (HTML '  </tr>' . "\n");

	#
	# search and output stats
	#
	@list = <NAMES>;
	@mdata = <MONTH>;
	&unlock(NAMES);
	close(NAMES);
	&unlock(MONTH);
	close(MONTH);

	foreach $name (@list) {
		chomp($name);
		@words = split(/\|/, $name);
		if ($pattern =~ /^$words[0]$/) {
			@words = split(/\|/, $name);
			$company = $words[2];
			@stuff = split(/\//, $words[0]);
			$ipstart = $stuff[0];
			@stuff = split(/\//, $words[1]);
			$ipend = $stuff[0];
			#
			# create ip range and zero totals
			#
			if ($ipstart =~ /$INTERFACE/) {
				$count = 0;
				$totalin = 0;
				$totalout = 0;
			}
			else {
				@stuff = split(/\./, $ipstart);
				@stuff2 = split(/\./, $ipend);
				$count = $stuff2[3] - $stuff[3];
				$totalin = 0;
				$totalout = 0;
			}
	
			#
			# parse database for ip range and output stats for individual ip
			#

			$tempcount = $count; 

			while ($count >= 0) {
				foreach $line (@mdata) {
					chomp($line);
					@words = split(/\|/, $line);
					if ($words[0] eq $ipstart) {
						$go = 1;
					}
					if ($go && ($count >= 0)) {
						$count--;
						$ip = $words[0];
						$totalin += $words[1];
						$totalout += $words[2];
						$mbytein = ($words[1] / 1024) / 1024;
						$mbyteout = ($words[2] / 1024) / 1024;
						$gbytein = $mbytein / 1024;
						$gbyteout = $mbyteout / 1024;
						print (HTML '  <tr>' . "\n");
						print (HTML '    <td align="center"><b>' . $company . '</b> [' . $ip . '] </td>' . "\n");
						printf (HTML "    <td align=\"center\"> %.3f &nbsp; %.3f </td>\n", $mbytein, $gbytein);
						printf (HTML "    <td align=\"center\"> %.3f &nbsp; %.3f </td>\n", $mbyteout, $gbyteout);
						print (HTML '  </tr>' . "\n");
					}
					else {
						break;
					}
				}

				if ($tempcount == $count) {
					#not found in file
					$count = -2;
				}
			}
			$go = 0;
		
			#
			# print totals for company
			#

			if ($count != -2) {
				print (HTML '  <tr bgcolor="#abcdef">' . "\n");
				print (HTML '    <td bgcolor="#abcdef"><b><u> Total [' . $company . ']: </u></b></td>' . "\n");
				printf (HTML "    <td bgcolor=\"#abcdef\" align=\"center\"> %.3f &nbsp; %.3f </td>\n", ($totalin/1024/1024), ($totalin/1024/1024/1024));
				printf (HTML "    <td bgcolor=\"#abcdef\" align=\"center\"> %.3f &nbsp; %.3f </td>\n", ($totalout/1024/1024), ($totalout/1024/1024/1024));
				print (HTML '  </tr>' . "\n");
			}
		}
	}
	
	#
	# print footers
	#
	print (HTML '  </table>' . "\n");
	print (HTML '</body>' . "\n");
	print (HTML '</html>' . "\n");
	&unlock(HTML);
	close(HTML);
}

#
# collects ipfilter stats for input and output
#
sub gather {
	open(DAYIN, ">$dayin") || die("Unable to open $dayin: $!\n");
	&lock(DAYIN);
	open(DAYOUT, ">$dayout") || die("Unable to open $dayout: $!\n");
	&lock(DAYOUT);
	@data = `ipfstat -ia`;
	@data2 = `ipfstat -oa`;
	
	#
	# read data from input array, test if it is an interface and output it
	#
	foreach $line (@data) {
		chomp($line);
		@words = split(/ +/, $line);
		$stats = $words[0];
		if ($words[6] =~ /any/) {
			@stuff = split(/\//, $words[8]);
			$ip = $stuff[0];
		}
		else {
			@stuff = split(/\//, $words[6]);
			$ip = $stuff[0];
		}
		if ($stats < 0) {
			$stats = 0;
		}
		print (DAYIN "$ip|$stats\n");
	}
	&unlock(DAYIN);
	close(DAYIN);
	
	#
	# read data from output array and output it
	#
	foreach $line (@data2) {
		chomp($line);
		@words = split(/ +/, $line);
		$stats = $words[0];
		if ($words[8] =~ /any/) {
			@stuff = split(/\//, $words[6]);
			$ip = $stuff[0];
		}
		else {
			@stuff = split(/\//, $words[8]);
			$ip = $stuff[0];
		}
		if ($stats < 0) {
			$stats = 0;
		}
		print (DAYOUT "$ip|$stats\n");
	}
	&unlock(DAYOUT);
	close(DAYOUT);
}

#
# converge daily input and output data from ipfilter
#
sub converge {
	open(DAYIN, "<$dayin") || die("Unable to open $dayin: $!\n");
	&lock(DAYIN);
	open(DAYOUT, "<$dayout") || die("Unable to open $dayout: $!\n");
	&lock(DAYOUT);
	open(DAY, ">$day") || die("Unable to open $day: $!\n");
	&lock(DAY);
	@data = <DAYIN>;
	@data2 = <DAYOUT>;
	
	#
	# clean up
	#
	&unlock(DAYIN);
	close(DAYIN);
	unlink $dayin;
	&unlock(DAYOUT);
	close(DAYOUT);
	unlink $dayout;
	
	#
	# read data from array and extract byte statistics
	#
	foreach $line (@data) {
		chomp($line);
		@words = split(/\|/, $line);
		$ip = $words[0];
		$stats = $words[1];
		foreach $line2 (@data2) {
			chomp($line2);
			@words2 = split(/\|/, $line2);
			if ($words2[0] eq $ip) {
				print (DAY "$ip|$stats|$words2[1]\n");
				break;
			}
		}
	}
	&unlock(DAY);
	close(DAY);
}

#
# compile daily data into monthly database
#
sub compile {
	open(DAY, "<$day") || die("Unable to open $day: $!\n");
	&lock(DAY);
	if (-e $month) {
		open(MONTH, "<$month") || die("Unable to open $month: $!\n");
		&lock(MONTH);
	}
	else {
		open(MONTH, ">$month") || die("Unable to open $month: $!\n");
		&lock(MONTH);
		print (MONTH "Start Date: $currentdate\n");
		print (MONTH "Report Date: $currentdate\n");
		&unlock(MONTH);
		close(MONTH);
		open(MONTH, "<$month") || die("Unable to open $month: $!\n");
		&lock(MONTH);
	}
	open(TEMP, ">$temp") || die("Unable to open $temp: $!\n");
	&lock(TEMP);
	
	#
	# handle date headers
	#
	chomp ($line = <MONTH>);
	print (TEMP "$line\n");
	chomp ($line = <MONTH>);
	print (TEMP "Report Date: $currentdate\n");

	#
	# read, parse and output
	#
	@data = <MONTH>;
	&unlock(MONTH);
	close(MONTH);
	while (defined($line = <DAY>)) {
		chomp($line);
		@words = split(/\|/, $line);
		$ip = $words[0];
		$statsin = $words[1];
		$statsout = $words[2];
		if ($statsin < 0) {
			$statsin = 0;
		}
		if ($statsout < 0) {
			$statsout = 0;
		}
		$stats2in = 0;
		$stats2out = 0;
		foreach $line (@data) {
			chomp($line);
			@words = split(/\|/, $line);
			if ($ip =~ /^$words[0]$/) {
				@words = split(/\|/, $line);
				$stats2in = $words[1];
				$stats2out = $words[2];
				break;
			}
		}
		print (TEMP "$ip|" . ($statsin + $stats2in) . "|" . ($statsout + $stats2out) . "\n");
	}
	&unlock(DAY);
	close(DAY);
	unlink $day;
	&unlock(TEMP);
	close(TEMP);
	rename($temp, $month);
}

#
# create webpage output
#
sub webify {
	open(MONTH, "<$month") || die("Unable to open $month: $!\n");
	&lock(MONTH);
	open(HTML, ">$html") || die("Unable to open $html: $!\n");
	&lock(HTML);
	open(NAMES, "<$names") || die("Unable to open $names: $!\n");
	&lock(NAMES);
	
	#
	# print headers
	#
	print (HTML '<html>' . "\n");
	print (HTML '<head>' . "\n");
	print (HTML '  <title>IP Accounting Information</title>' . "\n");
	print (HTML '</head>' . "\n");
	print (HTML '<body bgcolor="#ffffff">' . "\n");

	#
	# handle date headers
	#
	$line = <MONTH>;
	print (HTML $line);
	$line = <MONTH>;
	print (HTML "\nReport Date: $currentdate\n");
	
	#
	# begin table
	#
	print (HTML '<table border="0" width="90%">' . "\n");
	print (HTML '  <tr>' . "\n");
	print (HTML '    <td align="center"><b><u> IP address </u></b></td>' . "\n");
	print (HTML '    <td align="center"><b><u> MBytes in | GBytes in </u></b></td>' . "\n");
	print (HTML '    <td align="center"><b><u> MBytes out | GBytes out </u></b></td>' . "\n");
	print (HTML '  </tr>' . "\n");

	#
	# search and output stats
	#
	@list = <NAMES>;
	@mdata = <MONTH>;
	&unlock(NAMES);
	close(NAMES);
	&unlock(MONTH);
	close(MONTH);

	foreach $name (@list) {
		chomp($name);
		@words = split(/\|/, $name);
		$company = $words[2];
		@stuff = split(/\//, $words[0]);
		$ipstart = $stuff[0];
		@stuff = split(/\//, $words[1]);
		$ipend = $stuff[0];
		
		#
		# create ip range and zero totals
		#
		if ($ipstart =~ /$INTERFACE/) {
			$count = 0;
			$totalin = 0;
			$totalout = 0;
		}
		else {
			@stuff = split(/\./, $ipstart);
			@stuff2 = split(/\./, $ipend);
			$count = $stuff2[3] - $stuff[3];
			$totalin = 0;
			$totalout = 0;
		}
		
		#
		# parse database for ip range and output stats for individual ip
		#

		$tempcount = $count;

		while ($count >= 0) {
			foreach $line (@mdata) {
				chomp($line);
				@words = split(/\|/, $line);
				if ($ipstart =~ /^$words[0]$/) {
					$go = 1;
				}
				if ($go && ($count >= 0)) {
					$count--;
					@words = split(/\|/, $line);
					$ip = $words[0];
					$totalin += $words[1];
					$totalout += $words[2];
					$mbytein = ($words[1] / 1024) / 1024;
					$mbyteout = ($words[2] / 1024) / 1024;
					$gbytein = $mbytein / 1024;
					$gbyteout = $mbyteout / 1024;
					print (HTML '  <tr>' . "\n");
					print (HTML '    <td align="center"><b>' . $company . '</b> [' . $ip . '] </td>' . "\n");
					printf (HTML "    <td align=\"center\"> %.3f &nbsp; %.3f </td>\n", $mbytein, $gbytein);
					printf (HTML "    <td align=\"center\"> %.3f &nbsp; %.3f </td>\n", $mbyteout, $gbyteout);
					print (HTML '  </tr>' . "\n");
				}
			}

			if ($tempcount == $count) {
				#not found in file
				$count = -2;
			}
		}
		$go = 0;

		#
		# print totals for company
		#

		if ($count != -2) {
			print (HTML '  <tr bgcolor="#abcdef">' . "\n");
			print (HTML '    <td bgcolor="#abcdef"><b><u> Total [' . $company . ']: </u></b></td>' . "\n");
			printf (HTML "    <td bgcolor=\"#abcdef\" align=\"center\"> %.3f &nbsp; %.3f </td>\n", ($totalin/1024/1024), ($totalin/1024/1024/1024));
			printf (HTML "    <td bgcolor=\"#abcdef\" align=\"center\"> %.3f &nbsp; %.3f </td>\n", ($totalout/1024/1024), ($totalout/1024/1024/1024));
			print (HTML '  </tr>' . "\n");
		}
	}

	#
	# print footers
	#
	print (HTML '  </table>' . "\n");
	print (HTML '</body>' . "\n");
	print (HTML '</html>' . "\n");
	&unlock(HTML);
	close(HTML);
}

#
# resets counters
#
sub resetcount {
	system("ipf -Fa -f $IPF_LOC");
}

#
# main
#
if ($ARGV[0] eq '-s') {

	if ($ARGV[2]) {
		$html = $ARGV[2];
	}
	if ($ARGV[3]) {
		$month = $ARGV[3];
	}

	if ($ARGV[1] eq '') {
		&usage();
	}
	else {
		&search($ARGV[1]);
	}
}

elsif ($ARGV[0] eq '-w') {
	if ($ARGV[1]) {
		$html = $ARGV[1];
	}
	if ($ARGV[2]) {
		$month = $ARGV[2];
	}
	&webify();
}
elsif ($ARGV[0] eq '-r') {
	&resetcount();
}
elsif ($ARGV[0] eq '-c') {
	&gather();
	&converge();
	&compile();
	&webify();
	&resetcount();
}
else {
	&usage();
}
